/*
 * Copyright 2010-2017, Tarantool AUTHORS, please see AUTHORS file.
 *
 * Redistribution and use in source and binary forms, with or
 * without modification, are permitted provided that the following
 * conditions are met:
 *
 * 1. Redistributions of source code must retain the above
 *    copyright notice, this list of conditions and the
 *    following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials
 *    provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY <COPYRIGHT HOLDER> ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * <COPYRIGHT HOLDER> OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * This is the implementation of the page cache subsystem or "pager".
 *
 * The pager is used to access a database disk file.  It implements
 * atomic commit and rollback through the use of a journal file that
 * is separate from the database file.  The pager also implements file
 * locking to prevent two processes from writing the same database
 * file simultaneously, or one process from reading the database while
 * another is writing.
 */
#ifndef SQLITE_OMIT_DISKIO
#include "sqliteInt.h"

/******************* NOTES ON THE DESIGN OF THE PAGER ************************
 *
 * This comment block describes invariants that hold when using a rollback
 * journal.  These invariants do not apply for journal_mode=WAL,
 * journal_mode=MEMORY, or journal_mode=OFF.
 *
 * Within this comment block, a page is deemed to have been synced
 * automatically as soon as it is written when PRAGMA synchronous=OFF.
 * Otherwise, the page is not synced until the xSync method of the VFS
 * is called successfully on the file containing the page.
 *
 * Definition:  A page of the database file is said to be "overwriteable" if
 * one or more of the following are true about the page:
 *
 *     (a)  The original content of the page as it was at the beginning of
 *          the transaction has been written into the rollback journal and
 *          synced.
 *
 *     (b)  The page was a freelist leaf page at the start of the transaction.
 *
 *     (c)  The page number is greater than the largest page that existed in
 *          the database file at the start of the transaction.
 *
 * (1) A page of the database file is never overwritten unless one of the
 *     following are true:
 *
 *     (a) The page and all other pages on the same sector are overwriteable.
 *
 *     (b) The atomic page write optimization is enabled, and the entire
 *         transaction other than the update of the transaction sequence
 *         number consists of a single page change.
 *
 * (2) The content of a page written into the rollback journal exactly matches
 *     both the content in the database when the rollback journal was written
 *     and the content in the database at the beginning of the current
 *     transaction.
 *
 * (3) Writes to the database file are an integer multiple of the page size
 *     in length and are aligned on a page boundary.
 *
 * (4) Reads from the database file are either aligned on a page boundary and
 *     an integer multiple of the page size in length or are taken from the
 *     first 100 bytes of the database file.
 *
 * (5) All writes to the database file are synced prior to the rollback journal
 *     being deleted, truncated, or zeroed.
 *
 * (6) If a master journal file is used, then all writes to the database file
 *     are synced prior to the master journal being deleted.
 *
 * Definition: Two databases (or the same database at two points it time)
 * are said to be "logically equivalent" if they give the same answer to
 * all queries.  Note in particular the content of freelist leaf
 * pages can be changed arbitrarily without affecting the logical equivalence
 * of the database.
 *
 * (7) At any time, if any subset, including the empty set and the total set,
 *     of the unsynced changes to a rollback journal are removed and the
 *     journal is rolled back, the resulting database file will be logically
 *     equivalent to the database file at the beginning of the transaction.
 *
 * (8) When a transaction is rolled back, the xTruncate method of the VFS
 *     is called to restore the database file to the same size it was at
 *     the beginning of the transaction.  (In some VFSes, the xTruncate
 *     method is a no-op, but that does not change the fact the SQLite will
 *     invoke it.)
 *
 * (9) Whenever the database file is modified, at least one bit in the range
 *     of bytes from 24 through 39 inclusive will be changed prior to releasing
 *     the EXCLUSIVE lock, thus signaling other connections on the same
 *     database to flush their caches.
 *
 * (10) The pattern of bits in bytes 24 through 39 shall not repeat in less
 *      than one billion transactions.
 *
 * (11) A database file is well-formed at the beginning and at the conclusion
 *      of every transaction.
 *
 * (12) An EXCLUSIVE lock is held on the database file when writing to
 *      the database file.
 *
 * (13) A SHARED lock is held on the database file while reading any
 *      content out of the database file.
 *
 *****************************************************************************/

/*
 * Macros for troubleshooting.  Normally turned off
 */
#if 0
int sqlite3PagerTrace = 1;	/* True to enable tracing */
#define sqlite3DebugPrintf printf
#define PAGERTRACE(X)     if( sqlite3PagerTrace ){ sqlite3DebugPrintf X; }
#else
#define PAGERTRACE(X)
#endif

/*
 * The following two macros are used within the PAGERTRACE() macros above
 * to print out file-descriptors.
 *
 * PAGERID() takes a pointer to a Pager struct as its argument. The
 * associated file-descriptor is returned. FILEHANDLEID() takes an sqlite3_file
 * struct as its argument.
 */
#define PAGERID(p) ((int)(p->fd))
#define FILEHANDLEID(fd) ((int)fd)

/*
 * The Pager.eState variable stores the current 'state' of a pager. A
 * pager may be in any one of the seven states shown in the following
 * state diagram.
 *
 *                            OPEN <------+------+
 *                              |         |      |
 *                              V         |      |
 *               +---------> READER-------+      |
 *               |              |                |
 *               |              V                |
 *               |<-------WRITER_LOCKED------> ERROR
 *               |              |                ^
 *               |              V                |
 *               |<------WRITER_CACHEMOD-------->|
 *               |              |                |
 *               |              V                |
 *               |<-------WRITER_DBMOD---------->|
 *               |              |                |
 *               |              V                |
 *               +<------WRITER_FINISHED-------->+
 *
 *
 * List of state transitions and the C [function] that performs each:
 *
 *   OPEN              -> READER              [sqlite3PagerSharedLock]
 *   READER            -> OPEN                [pager_unlock]
 *
 *   READER            -> WRITER_LOCKED       [sqlite3PagerBegin]
 *   WRITER_LOCKED     -> WRITER_CACHEMOD     [pager_open_journal]
 *   WRITER_CACHEMOD   -> WRITER_DBMOD        [syncJournal]
 *   WRITER_DBMOD      -> WRITER_FINISHED     [sqlite3PagerCommitPhaseOne]
 *   WRITER_***        -> READER              [pager_end_transaction]
 *
 *   WRITER_***        -> ERROR               [pager_error]
 *   ERROR             -> OPEN                [pager_unlock]
 *
 *
 *  OPEN:
 *
 *    The pager starts up in this state. Nothing is guaranteed in this
 *    state - the file may or may not be locked and the database size is
 *    unknown. The database may not be read or written.
 *
 *    * No read or write transaction is active.
 *    * Any lock, or no lock at all, may be held on the database file.
 *    * The dbSize, dbOrigSize and dbFileSize variables may not be trusted.
 *
 *  READER:
 *
 *    In this state all the requirements for reading the database in
 *    rollback (non-WAL) mode are met. Unless the pager is (or recently
 *    was) in exclusive-locking mode, a user-level read transaction is
 *    open. The database size is known in this state.
 *
 *    A connection running with locking_mode=normal enters this state when
 *    it opens a read-transaction on the database and returns to state
 *    OPEN after the read-transaction is completed. However a connection
 *    running in locking_mode=exclusive (including temp databases) remains in
 *    this state even after the read-transaction is closed. The only way
 *    a locking_mode=exclusive connection can transition from READER to OPEN
 *    is via the ERROR state (see below).
 *
 *    * A read transaction may be active (but a write-transaction cannot).
 *    * A SHARED or greater lock is held on the database file.
 *    * The dbSize variable may be trusted (even if a user-level read
 *      transaction is not active). The dbOrigSize and dbFileSize variables
 *      may not be trusted at this point.
 *    * If the database is a WAL database, then the WAL connection is open.
 *    * Even if a read-transaction is not open, it is guaranteed that
 *      there is no hot-journal in the file-system.
 *
 *  WRITER_LOCKED:
 *
 *    The pager moves to this state from READER when a write-transaction
 *    is first opened on the database. In WRITER_LOCKED state, all locks
 *    required to start a write-transaction are held, but no actual
 *    modifications to the cache or database have taken place.
 *
 *    IN WAL mode, WalBeginWriteTransaction() is called to lock the log file.
 *    If the connection is running with locking_mode=exclusive, an attempt
 *    is made to obtain an EXCLUSIVE lock on the database file.
 *
 *    * A write transaction is active.
 *    * If the connection is open in rollback-mode, a RESERVED or greater
 *      lock is held on the database file.
 *    * If the connection is open in WAL-mode, a WAL write transaction
 *      is open (i.e. sqlite3WalBeginWriteTransaction() has been successfully
 *      called).
 *    * The dbSize, dbOrigSize and dbFileSize variables are all valid.
 *    * The contents of the pager cache have not been modified.
 *    * The journal file may or may not be open.
 *    * Nothing (not even the first header) has been written to the journal.
 *
 *  WRITER_CACHEMOD:
 *
 *    A pager moves from WRITER_LOCKED state to this state when a page is
 *    first modified by the upper layer. In rollback mode the journal file
 *    is opened (if it is not already open) and a header written to the
 *    start of it. The database file on disk has not been modified.
 *
 *    * A write transaction is active.
 *    * A RESERVED or greater lock is held on the database file.
 *    * The journal file is open and the first header has been written
 *      to it, but the header has not been synced to disk.
 *    * The contents of the page cache have been modified.
 *
 *  WRITER_DBMOD:
 *
 *    The pager transitions from WRITER_CACHEMOD into WRITER_DBMOD state
 *    when it modifies the contents of the database file. WAL connections
 *    never enter this state (since they do not modify the database file,
 *    just the log file).
 *
 *    * A write transaction is active.
 *    * An EXCLUSIVE or greater lock is held on the database file.
 *    * The journal file is open and the first header has been written
 *      and synced to disk.
 *    * The contents of the page cache have been modified (and possibly
 *      written to disk).
 *
 *  WRITER_FINISHED:
 *
 *    It is not possible for a WAL connection to enter this state.
 *
 *    A rollback-mode pager changes to WRITER_FINISHED state from WRITER_DBMOD
 *    state after the entire transaction has been successfully written into the
 *    database file. In this state the transaction may be committed simply
 *    by finalizing the journal file. Once in WRITER_FINISHED state, it is
 *    not possible to modify the database further. At this point, the upper
 *    layer must either commit or rollback the transaction.
 *
 *    * A write transaction is active.
 *    * An EXCLUSIVE or greater lock is held on the database file.
 *    * All writing and syncing of journal and database data has finished.
 *      If no error occurred, all that remains is to finalize the journal to
 *      commit the transaction. If an error did occur, the caller will need
 *      to rollback the transaction.
 *
 *  ERROR:
 *
 *    The ERROR state is entered when an IO or disk-full error (including
 *    SQLITE_IOERR_NOMEM) occurs at a point in the code that makes it
 *    difficult to be sure that the in-memory pager state (cache contents,
 *    db size etc.) are consistent with the contents of the file-system.
 *
 *    Temporary pager files may enter the ERROR state, but in-memory pagers
 *    cannot.
 *
 *    For example, if an IO error occurs while performing a rollback,
 *    the contents of the page-cache may be left in an inconsistent state.
 *    At this point it would be dangerous to change back to READER state
 *    (as usually happens after a rollback). Any subsequent readers might
 *    report database corruption (due to the inconsistent cache), and if
 *    they upgrade to writers, they may inadvertently corrupt the database
 *    file. To avoid this hazard, the pager switches into the ERROR state
 *    instead of READER following such an error.
 *
 *    Once it has entered the ERROR state, any attempt to use the pager
 *    to read or write data returns an error. Eventually, once all
 *    outstanding transactions have been abandoned, the pager is able to
 *    transition back to OPEN state, discarding the contents of the
 *    page-cache and any other in-memory state at the same time. Everything
 *    is reloaded from disk (and, if necessary, hot-journal rollback peformed)
 *    when a read-transaction is next opened on the pager (transitioning
 *    the pager into READER state). At that point the system has recovered
 *    from the error.
 *
 *    Specifically, the pager jumps into the ERROR state if:
 *
 *      1. An error occurs while attempting a rollback. This happens in
 *         function sqlite3PagerRollback().
 *
 *      2. An error occurs while attempting to finalize a journal file
 *         following a commit in function sqlite3PagerCommitPhaseTwo().
 *
 *      3. An error occurs while attempting to write to the journal or
 *         database file in function pagerStress() in order to free up
 *         memory.
 *
 *    In other cases, the error is returned to the b-tree layer. The b-tree
 *    layer then attempts a rollback operation. If the error condition
 *    persists, the pager enters the ERROR state via condition (1) above.
 *
 *    Condition (3) is necessary because it can be triggered by a read-only
 *    statement executed within a transaction. In this case, if the error
 *    code were simply returned to the user, the b-tree layer would not
 *    automatically attempt a rollback, as it assumes that an error in a
 *    read-only statement cannot leave the pager in an internally inconsistent
 *    state.
 *
 *    * The Pager.errCode variable is set to something other than SQLITE_OK.
 *    * There are one or more outstanding references to pages (after the
 *      last reference is dropped the pager should move back to OPEN state).
 *    * The pager is not an in-memory pager.
 *
 *
 * Notes:
 *
 *   * A pager is never in WRITER_DBMOD or WRITER_FINISHED state if the
 *     connection is open in WAL mode. A WAL connection is always in one
 *     of the first four states.
 *
 *   * Normally, a connection open in exclusive mode is never in PAGER_OPEN
 *     state. There are two exceptions: immediately after exclusive-mode has
 *     been turned on (and before any read or write transactions are
 *     executed), and when the pager is leaving the "error state".
 *
 *   * See also: assert_pager_state().
 */
#define PAGER_OPEN                  0
#define PAGER_READER                1
#define PAGER_WRITER_LOCKED         2
#define PAGER_WRITER_CACHEMOD       3
#define PAGER_WRITER_DBMOD          4
#define PAGER_WRITER_FINISHED       5
#define PAGER_ERROR                 6

/*
 * The Pager.eLock variable is almost always set to one of the
 * following locking-states, according to the lock currently held on
 * the database file: NO_LOCK, SHARED_LOCK, RESERVED_LOCK or EXCLUSIVE_LOCK.
 * This variable is kept up to date as locks are taken and released by
 * the pagerLockDb() and pagerUnlockDb() wrappers.
 *
 * If the VFS xLock() or xUnlock() returns an error other than SQLITE_BUSY
 * (i.e. one of the SQLITE_IOERR subtypes), it is not clear whether or not
 * the operation was successful. In these circumstances pagerLockDb() and
 * pagerUnlockDb() take a conservative approach - eLock is always updated
 * when unlocking the file, and only updated when locking the file if the
 * VFS call is successful. This way, the Pager.eLock variable may be set
 * to a less exclusive (lower) value than the lock that is actually held
 * at the system level, but it is never set to a more exclusive value.
 *
 * This is usually safe. If an xUnlock fails or appears to fail, there may
 * be a few redundant xLock() calls or a lock may be held for longer than
 * required, but nothing really goes wrong.
 *
 * The exception is when the database file is unlocked as the pager moves
 * from ERROR to OPEN state. At this point there may be a hot-journal file
 * in the file-system that needs to be rolled back (as part of an OPEN->SHARED
 * transition, by the same pager or any other). If the call to xUnlock()
 * fails at this point and the pager is left holding an EXCLUSIVE lock, this
 * can confuse the call to xCheckReservedLock() call made later as part
 * of hot-journal detection.
 *
 * xCheckReservedLock() is defined as returning true "if there is a RESERVED
 * lock held by this process or any others". So xCheckReservedLock may
 * return true because the caller itself is holding an EXCLUSIVE lock (but
 * doesn't know it because of a previous error in xUnlock). If this happens
 * a hot-journal may be mistaken for a journal being created by an active
 * transaction in another process, causing SQLite to read from the database
 * without rolling it back.
 *
 * To work around this, if a call to xUnlock() fails when unlocking the
 * database in the ERROR state, Pager.eLock is set to UNKNOWN_LOCK. It
 * is only changed back to a real locking state after a successful call
 * to xLock(EXCLUSIVE). Also, the code to do the OPEN->SHARED state transition
 * omits the check for a hot-journal if Pager.eLock is set to UNKNOWN_LOCK
 * lock. Instead, it assumes a hot-journal exists and obtains an EXCLUSIVE
 * lock on the database file before attempting to roll it back. See function
 * PagerSharedLock() for more detail.
 *
 * Pager.eLock may only be set to UNKNOWN_LOCK when the pager is in
 * PAGER_OPEN state.
 */
#define UNKNOWN_LOCK                (EXCLUSIVE_LOCK+1)

/*
 * A macro used for invoking the codec if there is one
 */
#ifdef SQLITE_HAS_CODEC
#define CODEC1(P,D,N,X,E) \
    if( P->xCodec && P->xCodec(P->pCodec,D,N,X)==0 ){ E; }
#define CODEC2(P,D,N,X,E,O) \
    if( P->xCodec==0 ){ O=(char*)D; }else \
    if( (O=(char*)(P->xCodec(P->pCodec,D,N,X)))==0 ){ E; }
#else
#define CODEC1(P,D,N,X,E)	/* NO-OP */
#define CODEC2(P,D,N,X,E,O) O=(char*)D
#endif

/*
 * The maximum allowed sector size. 64KiB. If the xSectorsize() method
 * returns a value larger than this, then MAX_SECTOR_SIZE is used instead.
 * This could conceivably cause corruption following a power failure on
 * such a system. This is currently an undocumented limit.
 */
#define MAX_SECTOR_SIZE 0x10000

/*
 * An instance of the following structure is allocated for each active
 * savepoint and statement transaction in the system. All such structures
 * are stored in the Pager.aSavepoint[] array, which is allocated and
 * resized using sqlite3Realloc().
 *
 * When a savepoint is created, the PagerSavepoint.iHdrOffset field is
 * set to 0. If a journal-header is written into the main journal while
 * the savepoint is active, then iHdrOffset is set to the byte offset
 * immediately following the last journal record written into the main
 * journal before the journal-header. This is required during savepoint
 * rollback (see pagerPlaybackSavepoint()).
 */
typedef struct PagerSavepoint PagerSavepoint;
struct PagerSavepoint {
	i64 iOffset;		/* Starting offset in main journal */
	i64 iHdrOffset;		/* See above */
	Bitvec *pInSavepoint;	/* Set of pages in this savepoint */
	Pgno nOrig;		/* Original number of pages in file */
	Pgno iSubRec;		/* Index of first record in sub-journal */
};

/*
 * Bits of the Pager.doNotSpill flag.  See further description below.
 */
#define SPILLFLAG_ROLLBACK    0x02	/* Current rolling back, so do not spill */

/*
 * An open page cache is an instance of struct Pager. A description of
 * some of the more important member variables follows:
 *
 * eState
 *
 *   The current 'state' of the pager object. See the comment and state
 *   diagram above for a description of the pager state.
 *
 * eLock
 *
 *   For a real on-disk database, the current lock held on the database file -
 *   NO_LOCK, SHARED_LOCK, RESERVED_LOCK or EXCLUSIVE_LOCK.
 *
 *   For a temporary or in-memory database (neither of which require any
 *   locks), this variable is always set to EXCLUSIVE_LOCK. Since such
 *   databases always have Pager.exclusiveMode==1, this tricks the pager
 *   logic into thinking that it already has all the locks it will ever
 *   need (and no reason to release them).
 *
 *   In some (obscure) circumstances, this variable may also be set to
 *   UNKNOWN_LOCK. See the comment above the #define of UNKNOWN_LOCK for
 *   details.
 *
 * changeCountDone
 *
 *   This boolean variable is used to make sure that the change-counter
 *   (the 4-byte header field at byte offset 24 of the database file) is
 *   not updated more often than necessary.
 *
 *   It is set to true when the change-counter field is updated, which
 *   can only happen if an exclusive lock is held on the database file.
 *   It is cleared (set to false) whenever an exclusive lock is
 *   relinquished on the database file. Each time a transaction is committed,
 *   The changeCountDone flag is inspected. If it is true, the work of
 *   updating the change-counter is omitted for the current transaction.
 *
 *   This mechanism means that when running in exclusive mode, a connection
 *   need only update the change-counter once, for the first transaction
 *   committed.
 *
 * setMaster
 *
 *   When PagerCommitPhaseOne() is called to commit a transaction, it may
 *   (or may not) specify a master-journal name to be written into the
 *   journal file before it is synced to disk.
 *
 *   Whether or not a journal file contains a master-journal pointer affects
 *   the way in which the journal file is finalized after the transaction is
 *   committed or rolled back when running in "journal_mode=PERSIST" mode.
 *   If a journal file does not contain a master-journal pointer, it is
 *   finalized by overwriting the first journal header with zeroes. If
 *   it does contain a master-journal pointer the journal file is finalized
 *   by truncating it to zero bytes, just as if the connection were
 *   running in "journal_mode=truncate" mode.
 *
 *   Journal files that contain master journal pointers cannot be finalized
 *   simply by overwriting the first journal-header with zeroes, as the
 *   master journal pointer could interfere with hot-journal rollback of any
 *   subsequently interrupted transaction that reuses the journal file.
 *
 *   The flag is cleared as soon as the journal file is finalized (either
 *   by PagerCommitPhaseTwo or PagerRollback). If an IO error prevents the
 *   journal file from being successfully finalized, the setMaster flag
 *   is cleared anyway (and the pager will move to ERROR state).
 *
 * doNotSpill
 *
 *   This variables control the behavior of cache-spills  (calls made by
 *   the pcache module to the pagerStress() routine to write cached data
 *   to the file-system in order to free up memory).
 *
 *   When bits SPILLFLAG_OFF or SPILLFLAG_ROLLBACK of doNotSpill are set,
 *   writing to the database from pagerStress() is disabled altogether.
 *   The SPILLFLAG_ROLLBACK case is done in a very obscure case that
 *   comes up during savepoint rollback that requires the pcache module
 *   to allocate a new page to prevent the journal file from being written
 *   while it is being traversed by code in pager_playback().  The SPILLFLAG_OFF
 *   case is a user preference.
 *
 *   If the SPILLFLAG_NOSYNC bit is set, writing to the database from
 *   pagerStress() is permitted, but syncing the journal file is not.
 *   This flag is set by sqlite3PagerWrite() when the file-system sector-size
 *   is larger than the database page-size in order to prevent a journal sync
 *   from happening in between the journalling of two pages on the same sector.
 *
 * subjInMemory
 *
 *   This is a boolean variable. If true, then any required sub-journal
 *   is opened as an in-memory journal file. If false, then in-memory
 *   sub-journals are only used for in-memory pager files.
 *
 *   This variable is updated by the upper layer each time a new
 *   write-transaction is opened.
 *
 * dbSize, dbOrigSize, dbFileSize
 *
 *   Variable dbSize is set to the number of pages in the database file.
 *   It is valid in PAGER_READER and higher states (all states except for
 *   OPEN and ERROR).
 *
 *   dbSize is set based on the size of the database file, which may be
 *   larger than the size of the database (the value stored at offset
 *   28 of the database header by the btree). If the size of the file
 *   is not an integer multiple of the page-size, the value stored in
 *   dbSize is rounded down (i.e. a 5KB file with 2K page-size has dbSize==2).
 *   Except, any file that is greater than 0 bytes in size is considered
 *   to have at least one page. (i.e. a 1KB file with 2K page-size leads
 *   to dbSize==1).
 *
 *   During a write-transaction, if pages with page-numbers greater than
 *   dbSize are modified in the cache, dbSize is updated accordingly.
 *   Similarly, if the database is truncated using PagerTruncateImage(),
 *   dbSize is updated.
 *
 *   Variables dbOrigSize and dbFileSize are valid in states
 *   PAGER_WRITER_LOCKED and higher. dbOrigSize is a copy of the dbSize
 *   variable at the start of the transaction. It is used during rollback,
 *   and to determine whether or not pages need to be journalled before
 *   being modified.
 *
 *   Throughout a write-transaction, dbFileSize contains the size of
 *   the file on disk in pages. It is set to a copy of dbSize when the
 *   write-transaction is first opened, and updated when VFS calls are made
 *   to write or truncate the database file on disk.
 *
 *   The only reason the dbFileSize variable is required is to suppress
 *   unnecessary calls to xTruncate() after committing a transaction. If,
 *   when a transaction is committed, the dbFileSize variable indicates
 *   that the database file is larger than the database image (Pager.dbSize),
 *   pager_truncate() is called. The pager_truncate() call uses xFilesize()
 *   to measure the database file on disk, and then truncates it if required.
 *   dbFileSize is not used when rolling back a transaction. In this case
 *   pager_truncate() is called unconditionally (which means there may be
 *   a call to xFilesize() that is not strictly required). In either case,
 *   pager_truncate() may cause the file to become smaller or larger.
 *
 * dbHintSize
 *
 *   The dbHintSize variable is used to limit the number of calls made to
 *   the VFS xFileControl(FCNTL_SIZE_HINT) method.
 *
 *   dbHintSize is set to a copy of the dbSize variable when a
 *   write-transaction is opened (at the same time as dbFileSize and
 *   dbOrigSize). If the xFileControl(FCNTL_SIZE_HINT) method is called,
 *   dbHintSize is increased to the number of pages that correspond to the
 *   size-hint passed to the method call. See pager_write_pagelist() for
 *   details.
 *
 * errCode
 *
 *   The Pager.errCode variable is only ever used in PAGER_ERROR state. It
 *   is set to zero in all other states. In PAGER_ERROR state, Pager.errCode
 *   is always set to SQLITE_FULL, SQLITE_IOERR or one of the SQLITE_IOERR_XXX
 *   sub-codes.
 */
struct Pager {
	sqlite3_vfs *pVfs;	/* OS functions to use for IO */
	u8 exclusiveMode;	/* Boolean. True if locking_mode==EXCLUSIVE */
	u8 journalMode;		/* One of the PAGER_JOURNALMODE_* values */
	u8 useJournal;		/* Use a rollback journal on this file */
	u8 noSync;		/* Do not sync the journal if true */
	u8 fullSync;		/* Do extra syncs of the journal for robustness */
	u8 extraSync;		/* sync directory after journal delete */
	u8 ckptSyncFlags;	/* SYNC_NORMAL or SYNC_FULL for checkpoint */
	u8 walSyncFlags;	/* SYNC_NORMAL or SYNC_FULL for wal writes */
	u8 syncFlags;		/* SYNC_NORMAL or SYNC_FULL otherwise */
	u8 tempFile;		/* zFilename is a temporary or immutable file */
	u8 noLock;		/* Do not lock (except in WAL mode) */
	u8 readOnly;		/* True for a read-only database */
	u8 memDb;		/* True to inhibit all file I/O */

  /**************************************************************************
  * The following block contains those class members that change during
  * routine operation.  Class members not in this block are either fixed
  * when the pager is first created or else only change when there is a
  * significant mode change (such as changing the page_size, locking_mode,
  * or the journal_mode).  From another view, these class members describe
  * the "state" of the pager, while other class members describe the
  * "configuration" of the pager.
  */
	u8 eState;		/* Pager state (OPEN, READER, WRITER_LOCKED..) */
	u8 eLock;		/* Current lock held on database file */
	u8 changeCountDone;	/* Set after incrementing the change-counter */
	u8 setMaster;		/* True if a m-j name has been written to jrnl */
	u8 doNotSpill;		/* Do not spill the cache when non-zero */
	u8 subjInMemory;	/* True to use in-memory sub-journals */
	u8 bUseFetch;		/* True to use xFetch() */
	u8 hasHeldSharedLock;	/* True if a shared lock has ever been held */
	Pgno dbSize;		/* Number of pages in the database */
	Pgno dbOrigSize;	/* dbSize before the current transaction */
	Pgno dbFileSize;	/* Number of pages in the database file */
	Pgno dbHintSize;	/* Value passed to FCNTL_SIZE_HINT call */
	int errCode;		/* One of several kinds of errors */
	int nRec;		/* Pages journalled since last j-header written */
	u32 cksumInit;		/* Quasi-random value added to every checksum */
	u32 nSubRec;		/* Number of records written to sub-journal */
	Bitvec *pInJournal;	/* One bit for each page in the database file */
	sqlite3_file *fd;	/* File descriptor for database */
	sqlite3_file *jfd;	/* File descriptor for main journal */
	sqlite3_file *sjfd;	/* File descriptor for sub-journal */
	i64 journalOff;		/* Current write offset in the journal file */
	i64 journalHdr;		/* Byte offset to previous journal header */
	PagerSavepoint *aSavepoint;	/* Array of active savepoints */
	int nSavepoint;		/* Number of elements in aSavepoint[] */
	u32 iDataVersion;	/* Changes whenever database content changes */
	char dbFileVers[16];	/* Changes whenever database file changes */

	int nMmapOut;		/* Number of mmap pages currently outstanding */
	sqlite3_int64 szMmap;	/* Desired maximum mmap size */
	PgHdr *pMmapFreelist;	/* List of free mmap page headers (pDirty) */
	/*
	 * End of the routinely-changing class members
	 **************************************************************************/

	u16 nExtra;		/* Add this many bytes to each in-memory page */
	i16 nReserve;		/* Number of unused bytes at end of each page */
	u32 vfsFlags;		/* Flags for sqlite3_vfs.xOpen() */
	u32 sectorSize;		/* Assumed sector size during rollback */
	int pageSize;		/* Number of bytes in a page */
	Pgno mxPgno;		/* Maximum allowed size of the database */
	i64 journalSizeLimit;	/* Size limit for persistent journal files */
	char *zFilename;	/* Name of the database file */
	char *zJournal;		/* Name of the journal file */
	int (*xBusyHandler) (void *);	/* Function to call when busy */
	void *pBusyHandlerArg;	/* Context argument for xBusyHandler */
	int aStat[3];		/* Total cache hits, misses and writes */
#ifdef SQLITE_TEST
	int nRead;		/* Database pages read */
#endif
	void (*xReiniter) (DbPage *);	/* Call this routine when reloading pages */
	int (*xGet) (Pager *, Pgno, DbPage **, int);	/* Routine to fetch a patch */
#ifdef SQLITE_HAS_CODEC
	void *(*xCodec) (void *, void *, Pgno, int);	/* Routine for en/decoding data */
	void (*xCodecSizeChng) (void *, int, int);	/* Notify of page size changes */
	void (*xCodecFree) (void *);	/* Destructor for the codec */
	void *pCodec;		/* First argument to xCodec... methods */
#endif
	char *pTmpSpace;	/* Pager.pageSize bytes of space for tmp use */
	PCache *pPCache;	/* Pointer to page cache object */
};

/*
 * Indexes for use with Pager.aStat[]. The Pager.aStat[] array contains
 * the values accessed by passing SQLITE_DBSTATUS_CACHE_HIT, CACHE_MISS
 * or CACHE_WRITE to sqlite3_db_status().
 */
#define PAGER_STAT_HIT   0

/*
 * The following global variables hold counters used for
 * testing purposes only.  These variables do not exist in
 * a non-testing build.  These variables are not thread-safe.
 */
#ifdef SQLITE_TEST
int sqlite3_pager_writej_count = 0;	/* Number of pages written to journal */
#define PAGER_INCR(v)  v++
#else
#define PAGER_INCR(v)
#endif

/*
 * The macro MEMDB is true if we are dealing with an in-memory database.
 * We do this as a macro so that if the SQLITE_OMIT_MEMORYDB macro is set,
 * the value of MEMDB will be a constant and the compiler will optimize
 * out code that would never execute.
 */
#ifdef SQLITE_OMIT_MEMORYDB
#define MEMDB 0
#else
#define MEMDB pPager->memDb
#endif

/*
 * The macro USEFETCH is true if we are allowed to use the xFetch and xUnfetch
 * interfaces to access the database using memory-mapped I/O.
 */
#if SQLITE_MAX_MMAP_SIZE>0
#define USEFETCH(x) ((x)->bUseFetch)
#else
#define USEFETCH(x) 0
#endif

/*
 * The maximum legal page number is (2^31 - 1).
 */
#define PAGER_MAX_PGNO 2147483647

/*
 * The argument to this macro is a file descriptor (type sqlite3_file*).
 * Return 0 if it is not open, or non-zero (but not 1) if it is.
 *
 * This is so that expressions can be written as:
 *
 *   if( isOpen(pPager->jfd) ){ ...
 *
 * instead of
 *
 *   if( pPager->jfd->pMethods ){ ...
 */
#define isOpen(pFd) ((pFd)->pMethods!=0)

#define pagerUseWal(x) 0
#define pagerBeginReadTransaction(z) SQLITE_OK

#ifndef NDEBUG
/*
 * Usage:
 *
 *   assert( assert_pager_state(pPager) );
 *
 * This function runs many asserts to try to find inconsistencies in
 * the internal state of the Pager object.
 */
static int
assert_pager_state(Pager * p)
{
	Pager *pPager = p;

	/* State must be valid. */
	assert(p->eState == PAGER_OPEN
	       || p->eState == PAGER_READER
	       || p->eState == PAGER_WRITER_LOCKED
	       || p->eState == PAGER_WRITER_CACHEMOD
	       || p->eState == PAGER_WRITER_DBMOD
	       || p->eState == PAGER_WRITER_FINISHED
	       || p->eState == PAGER_ERROR);

	/* Regardless of the current state, a temp-file connection always behaves
	 * as if it has an exclusive lock on the database file. It never updates
	 * the change-counter field, so the changeCountDone flag is always set.
	 */
	assert(p->tempFile == 0 || p->eLock == EXCLUSIVE_LOCK);
	assert(p->tempFile == 0 || pPager->changeCountDone);

	/* If the useJournal flag is clear, the journal-mode must be "OFF".
	 * And if the journal-mode is "OFF", the journal file must not be open.
	 */
	assert(p->journalMode == PAGER_JOURNALMODE_OFF || p->useJournal);
	assert(p->journalMode != PAGER_JOURNALMODE_OFF || !isOpen(p->jfd));

	/* Check that MEMDB implies noSync. And an in-memory journal. Since
	 * this means an in-memory pager performs no IO at all, it cannot encounter
	 * either SQLITE_IOERR or SQLITE_FULL during rollback or while finalizing
	 * a journal file. (although the in-memory journal implementation may
	 * return SQLITE_IOERR_NOMEM while the journal file is being written). It
	 * is therefore not possible for an in-memory pager to enter the ERROR
	 * state.
	 */
	if (MEMDB) {
		assert(!isOpen(p->fd));
		assert(p->noSync);
		assert(p->journalMode == PAGER_JOURNALMODE_OFF
		       || p->journalMode == PAGER_JOURNALMODE_MEMORY);
		assert(p->eState != PAGER_ERROR && p->eState != PAGER_OPEN);
		assert(pagerUseWal(p) == 0);
	}

	/* If changeCountDone is set, a RESERVED lock or greater must be held
	 * on the file.
	 */
	assert(pPager->changeCountDone == 0 || pPager->eLock >= RESERVED_LOCK);
	assert(p->eLock != PENDING_LOCK);

	switch (p->eState) {
	case PAGER_READER:
		assert(pPager->errCode == SQLITE_OK);
		assert(p->eLock != UNKNOWN_LOCK);
		assert(p->eLock >= SHARED_LOCK);
		break;

	case PAGER_WRITER_LOCKED:
		assert(p->eLock != UNKNOWN_LOCK);
		assert(pPager->errCode == SQLITE_OK);
		if (!pagerUseWal(pPager)) {
			assert(p->eLock >= RESERVED_LOCK);
		}
		assert(pPager->dbOrigSize == pPager->dbFileSize);
		assert(pPager->dbOrigSize == pPager->dbHintSize);
		assert(pPager->setMaster == 0);
		break;
	}

	return 1;
}
#endif				/* ifndef NDEBUG */

/* Forward references to the various page getters */
static int getPageNormal(Pager *, Pgno, DbPage **, int);
static int getPageError(Pager *, Pgno, DbPage **, int);

/*
 * Set the Pager.xGet method for the appropriate routine used to fetch
 * content from the pager.
 */
static void
setGetterMethod(Pager * pPager)
{
	if (pPager->errCode) {
		pPager->xGet = getPageError;
#if SQLITE_MAX_MMAP_SIZE>0
	} else if (USEFETCH(pPager)
#ifdef SQLITE_HAS_CODEC
		   && pPager->xCodec == 0
#endif
	    ) {
#endif				/* SQLITE_MAX_MMAP_SIZE>0 */
	} else {
		pPager->xGet = getPageNormal;
	}
}

/*
 * This function determines whether or not the atomic-write optimization
 * can be used with this pager. The optimization can be used if:
 *
 *  (a) the value returned by OsDeviceCharacteristics() indicates that
 *      a database page may be written atomically, and
 *  (b) the value returned by OsSectorSize() is less than or equal
 *      to the page size.
 *
 * The optimization is also always enabled for temporary files. It is
 * an error to call this function if pPager is opened on an in-memory
 * database.
 *
 * If the optimization cannot be used, 0 is returned. If it can be used,
 * then the value returned is the size of the journal file when it
 * contains rollback data for exactly one page.
 */
#ifdef SQLITE_ENABLE_ATOMIC_WRITE
static int
jrnlBufferSize(Pager * pPager)
{
	assert(!MEMDB);
	if (!pPager->tempFile) {
		int dc;		/* Device characteristics */
		int nSector;	/* Sector size */
		int szPage;	/* Page size */

		assert(isOpen(pPager->fd));
		dc = sqlite3OsDeviceCharacteristics(pPager->fd);
		nSector = pPager->sectorSize;
		szPage = pPager->pageSize;

		assert(SQLITE_IOCAP_ATOMIC512 == (512 >> 8));
		assert(SQLITE_IOCAP_ATOMIC64K == (65536 >> 8));
		if (0 ==
		    (dc & (SQLITE_IOCAP_ATOMIC | (szPage >> 8))
		     || nSector > szPage)) {
			return 0;
		}
	}

	return JOURNAL_HDR_SZ(pPager) + JOURNAL_PG_SZ(pPager);
}
#else
#define jrnlBufferSize(x) 0
#endif

/*
 * If SQLITE_CHECK_PAGES is defined then we do some sanity checking
 * on the cache using a hash function.  This is used for testing
 * and debugging only.
 */
#ifdef SQLITE_CHECK_PAGES
/*
 * Return a 32-bit hash of the page data for pPage.
 */
static u32
pager_datahash(int nByte, unsigned char *pData)
{
	u32 hash = 0;
	int i;
	for (i = 0; i < nByte; i++) {
		hash = (hash * 1039) + pData[i];
	}
	return hash;
}

static u32
pager_pagehash(PgHdr * pPage)
{
	return pager_datahash(pPage->pPager->pageSize,
			      (unsigned char *)pPage->pData);
}

static void
pager_set_pagehash(PgHdr * pPage)
{
	pPage->pageHash = pager_pagehash(pPage);
}

/*
 * The CHECK_PAGE macro takes a PgHdr* as an argument. If SQLITE_CHECK_PAGES
 * is defined, and NDEBUG is not defined, an assert() statement checks
 * that the page is either dirty or still matches the calculated page-hash.
 */
#define CHECK_PAGE(x) checkPage(x)
static void
checkPage(PgHdr * pPg)
{
	Pager *pPager = pPg->pPager;
	assert(pPager->eState != PAGER_ERROR);
	assert((pPg->flags & PGHDR_DIRTY)
	       || pPg->pageHash == pager_pagehash(pPg));
}

#else
#define pager_datahash(X,Y)  0
#define pager_pagehash(X)  0
#define pager_set_pagehash(X)
#define CHECK_PAGE(x)
#endif				/* SQLITE_CHECK_PAGES */

/*
 * Discard the entire contents of the in-memory page-cache.
 */
static void
pager_reset(Pager * pPager)
{
	pPager->iDataVersion++;
	sqlite3PcacheClear(pPager->pPCache);
}

/*
 * Return the pPager->iDataVersion value
 */
u32
sqlite3PagerDataVersion(Pager * pPager)
{
	assert(pPager->eState > PAGER_OPEN);
	return pPager->iDataVersion;
}

/*
 * Report the current page size and number of reserved bytes back
 * to the codec.
 */
#ifdef SQLITE_HAS_CODEC
static void
pagerReportSize(Pager * pPager)
{
	if (pPager->xCodecSizeChng) {
		pPager->xCodecSizeChng(pPager->pCodec, pPager->pageSize,
				       (int)pPager->nReserve);
	}
}
#else
#define pagerReportSize(X)	/* No-op if we do not support a codec */
#endif

/*
 * Set the value of the Pager.sectorSize variable for the given
 * pager based on the value returned by the xSectorSize method
 * of the open database file. The sector size will be used
 * to determine the size and alignment of journal header and
 * master journal pointers within created journal files.
 *
 * For temporary files the effective sector size is always 512 bytes.
 *
 * Otherwise, for non-temporary files, the effective sector size is
 * the value returned by the xSectorSize() method rounded up to 32 if
 * it is less than 32, or rounded down to MAX_SECTOR_SIZE if it
 * is greater than MAX_SECTOR_SIZE.
 *
 * If the file has the SQLITE_IOCAP_POWERSAFE_OVERWRITE property, then set
 * the effective sector size to its minimum value (512).  The purpose of
 * pPager->sectorSize is to define the "blast radius" of bytes that
 * might change if a crash occurs while writing to a single byte in
 * that range.  But with POWERSAFE_OVERWRITE, the blast radius is zero
 * (that is what POWERSAFE_OVERWRITE means), so we minimize the sector
 * size.  For backwards compatibility of the rollback journal file format,
 * we cannot reduce the effective sector size below 512.
 */
static void
setSectorSize(Pager * pPager)
{
	assert(isOpen(pPager->fd) || pPager->tempFile);

	if (pPager->tempFile
	    || (sqlite3OsDeviceCharacteristics(pPager->fd) &
		SQLITE_IOCAP_POWERSAFE_OVERWRITE) != 0) {
		/* Sector size doesn't matter for temporary files. Also, the file
		 * may not have been opened yet, in which case the OsSectorSize()
		 * call will segfault.
		 */
		pPager->sectorSize = 512;
	}
}

/*
 * Change the maximum number of in-memory pages that are allowed
 * before attempting to recycle clean and unused pages.
 */
void
sqlite3PagerSetCachesize(Pager * pPager, int mxPage)
{
	sqlite3PcacheSetCachesize(pPager->pPCache, mxPage);
}

/*
 * Change the page size used by the Pager object. The new page size
 * is passed in *pPageSize.
 *
 * If the pager is in the error state when this function is called, it
 * is a no-op. The value returned is the error state error code (i.e.
 * one of SQLITE_IOERR, an SQLITE_IOERR_xxx sub-code or SQLITE_FULL).
 *
 * Otherwise, if all of the following are true:
 *
 *   * the new page size (value of *pPageSize) is valid (a power
 *     of two between 512 and SQLITE_MAX_PAGE_SIZE, inclusive), and
 *
 *   * there are no outstanding page references, and
 *
 *   * the database is either not an in-memory database or it is
 *     an in-memory database that currently consists of zero pages.
 *
 * then the pager object page size is set to *pPageSize.
 *
 * If the page size is changed, then this function uses sqlite3PagerMalloc()
 * to obtain a new Pager.pTmpSpace buffer. If this allocation attempt
 * fails, SQLITE_NOMEM is returned and the page size remains unchanged.
 * In all other cases, SQLITE_OK is returned.
 *
 * If the page size is not changed, either because one of the enumerated
 * conditions above is not true, the pager was in error state when this
 * function was called, or because the memory allocation attempt failed,
 * then *pPageSize is set to the old, retained page size before returning.
 */
int
sqlite3PagerSetPagesize(Pager * pPager, u32 * pPageSize, int nReserve)
{
	int rc = SQLITE_OK;

	/* It is not possible to do a full assert_pager_state() here, as this
	 * function may be called from within PagerOpen(), before the state
	 * of the Pager object is internally consistent.
	 *
	 * At one point this function returned an error if the pager was in
	 * PAGER_ERROR state. But since PAGER_ERROR state guarantees that
	 * there is at least one outstanding page reference, this function
	 * is a no-op for that case anyhow.
	 */

	u32 pageSize = *pPageSize;
	assert(pageSize == 0
	       || (pageSize >= 512 && pageSize <= SQLITE_MAX_PAGE_SIZE));
	if ((pPager->memDb == 0 || pPager->dbSize == 0)
	    && sqlite3PcacheRefCount(pPager->pPCache) == 0
	    && pageSize && pageSize != (u32) pPager->pageSize) {
		char *pNew = NULL;	/* New temp space */
		i64 nByte = 0;

		if (pPager->eState > PAGER_OPEN && isOpen(pPager->fd)) {
			rc = sqlite3OsFileSize(pPager->fd, &nByte);
		}
		if (rc == SQLITE_OK) {
			pNew = (char *)sqlite3PageMalloc(pageSize);
			if (!pNew)
				rc = SQLITE_NOMEM_BKPT;
		}

		if (rc == SQLITE_OK) {
			pager_reset(pPager);
			rc = sqlite3PcacheSetPageSize(pPager->pPCache,
						      pageSize);
		}
		if (rc == SQLITE_OK) {
			sqlite3PageFree(pPager->pTmpSpace);
			pPager->pTmpSpace = pNew;
			pPager->dbSize =
			    (Pgno) ((nByte + pageSize - 1) / pageSize);
			pPager->pageSize = pageSize;
		} else {
			sqlite3PageFree(pNew);
		}
	}

	*pPageSize = pPager->pageSize;
	if (rc == SQLITE_OK) {
		if (nReserve < 0)
			nReserve = pPager->nReserve;
		assert(nReserve >= 0 && nReserve < 1000);
		pPager->nReserve = (i16) nReserve;
		pagerReportSize(pPager);
	}
	return rc;
}

/*
 * Return a pointer to the "temporary page" buffer held internally
 * by the pager.  This is a buffer that is big enough to hold the
 * entire content of a database page.  This buffer is used internally
 * during rollback and will be overwritten whenever a rollback
 * occurs.  But other modules are free to use it too, as long as
 * no rollbacks are happening.
 */
void *
sqlite3PagerTempSpace(Pager * pPager)
{
	return pPager->pTmpSpace;
}

/*
 * The following set of routines are used to disable the simulated
 * I/O error mechanism.  These routines are used to avoid simulated
 * errors in places where we do not care about errors.
 *
 * Unless -DSQLITE_TEST=1 is used, these routines are all no-ops
 * and generate no code.
 */
#ifdef SQLITE_TEST
extern int sqlite3_io_error_pending;
extern int sqlite3_io_error_hit;
static int saved_cnt;
void
disable_simulated_io_errors(void)
{
	saved_cnt = sqlite3_io_error_pending;
	sqlite3_io_error_pending = -1;
}

void
enable_simulated_io_errors(void)
{
	sqlite3_io_error_pending = saved_cnt;
}
#else
#define disable_simulated_io_errors()
#define enable_simulated_io_errors()
#endif

/*
 * Read the first N bytes from the beginning of the file into memory
 * that pDest points to.
 *
 * If the pager was opened on a transient file (zFilename==""), or
 * opened on a file less than N bytes in size, the output buffer is
 * zeroed and SQLITE_OK returned. The rationale for this is that this
 * function is used to read database headers, and a new transient or
 * zero sized database has a header than consists entirely of zeroes.
 *
 * If any IO error apart from SQLITE_IOERR_SHORT_READ is encountered,
 * the error code is returned to the caller and the contents of the
 * output buffer undefined.
 */
int
sqlite3PagerReadFileheader(Pager * pPager, int N, unsigned char *pDest)
{
	int rc = SQLITE_OK;
	memset(pDest, 0, N);
	assert(isOpen(pPager->fd) || pPager->tempFile);

	/* This routine is only called by btree immediately after creating
	 * the Pager object.  There has not been an opportunity to transition
	 * to WAL mode yet.
	 */
	assert(!pagerUseWal(pPager));
	return rc;
}

/*
 * This function may only be called when a read-transaction is open on
 * the pager. It returns the total number of pages in the database.
 *
 * However, if the file is between 1 and <page-size> bytes in size, then
 * this is considered a 1 page file.
 */
void
sqlite3PagerPagecount(Pager * pPager, int *pnPage)
{
	assert(pPager->eState >= PAGER_READER);
	assert(pPager->eState != PAGER_WRITER_FINISHED);
	*pnPage = (int)pPager->dbSize;
}

/*
 * Shutdown the page cache.  Free all memory and close all files.
 *
 * If a transaction was in progress when this routine is called, that
 * transaction is rolled back.  All outstanding pages are invalidated
 * and their memory is freed.  Any attempt to use a page associated
 * with this page cache after this function returns will likely
 * result in a coredump.
 *
 * This function always succeeds. If a transaction is active an attempt
 * is made to roll it back. If an error occurs during the rollback
 * a hot journal may be left in the filesystem but no error is returned
 * to the caller.
 */
int
sqlite3PagerClose(Pager * pPager, sqlite3 * db)
{
	u8 *pTmp = (u8 *) pPager->pTmpSpace;

	assert(db || pagerUseWal(pPager) == 0);
	assert(assert_pager_state(pPager));
	disable_simulated_io_errors();
	sqlite3BeginBenignMalloc();
	/* pPager->errCode = 0; */
	pPager->exclusiveMode = 0;
	pager_reset(pPager);
	sqlite3EndBenignMalloc();
	enable_simulated_io_errors();
	PAGERTRACE(("CLOSE %d\n", PAGERID(pPager)));
	IOTRACE(("CLOSE %p\n", pPager))
	    sqlite3OsClose(pPager->jfd);
	sqlite3OsClose(pPager->fd);
	sqlite3PageFree(pTmp);
	sqlite3PcacheClose(pPager->pPCache);

#ifdef SQLITE_HAS_CODEC
	if (pPager->xCodecFree)
		pPager->xCodecFree(pPager->pCodec);
#endif

	assert(!pPager->aSavepoint && !pPager->pInJournal);
	assert(!isOpen(pPager->jfd) && !isOpen(pPager->sjfd));

	sqlite3_free(pPager);
	return SQLITE_OK;
}

#if !defined(NDEBUG) || defined(SQLITE_TEST)
/*
 * Return the page number for page pPg.
 */
Pgno
sqlite3PagerPagenumber(DbPage * pPg)
{
	return pPg->pgno;
}
#endif

/*
 * Increment the reference count for page pPg.
 */
void
sqlite3PagerRef(DbPage * pPg)
{
	sqlite3PcacheRef(pPg);
}

/*
 * Allocate and initialize a new Pager object and put a pointer to it
 * in *ppPager. The pager should eventually be freed by passing it
 * to sqlite3PagerClose().
 *
 * The zFilename argument is the path to the database file to open.
 * If zFilename is NULL then a randomly-named temporary file is created
 * and used as the file to be cached. Temporary files are be deleted
 * automatically when they are closed. If zFilename is ":memory:" then
 * all information is held in cache. It is never written to disk.
 * This can be used to implement an in-memory database.
 *
 * The nExtra parameter specifies the number of bytes of space allocated
 * along with each page reference. This space is available to the user
 * via the sqlite3PagerGetExtra() API.  When a new page is allocated, the
 * first 8 bytes of this space are zeroed but the remainder is uninitialized.
 * (The extra space is used by btree as the MemPage object.)
 *
 * The flags argument is used to specify properties that affect the
 * operation of the pager. It should be passed some bitwise combination
 * of the PAGER_* flags.
 *
 * The vfsFlags parameter is a bitmask to pass to the flags parameter
 * of the xOpen() method of the supplied VFS when opening files.
 *
 * If the pager object is allocated and the specified file opened
 * successfully, SQLITE_OK is returned and *ppPager set to point to
 * the new pager object. If an error occurs, *ppPager is set to NULL
 * and error code returned. This function may return SQLITE_NOMEM
 * (sqlite3Malloc() is used to allocate memory), SQLITE_CANTOPEN or
 * various SQLITE_IO_XXX errors.
 */
int
sqlite3PagerOpen(sqlite3_vfs * pVfs,	/* The virtual file system to use */
		 Pager ** ppPager,	/* OUT: Return the Pager structure here */
		 const char *zFilename,	/* Name of the database file to open */
		 int nExtra,	/* Extra bytes append to each in-memory page */
		 int flags,	/* flags controlling this file */
		 int vfsFlags,	/* flags passed through to sqlite3_vfs.xOpen() */
		 void (*xReinit) (DbPage *)	/* Function to reinitialize pages */
    )
{
	u8 *pPtr;
	Pager *pPager = 0;	/* Pager object to allocate and return */
	int rc = SQLITE_OK;	/* Return code */
	int tempFile = 0;	/* True for temp files (incl. in-memory files) */
	int memDb = 0;		/* True if this is an in-memory file */
	int readOnly = 0;	/* True if this is a read-only file */
	int journalFileSize;	/* Bytes to allocate for each journal fd */
	char *zPathname = 0;	/* Full path to database file */
	int nPathname = 0;	/* Number of bytes in zPathname */
	int useJournal = (flags & PAGER_OMIT_JOURNAL) == 0;	/* False to omit journal */
	int pcacheSize = sqlite3PcacheSize();	/* Bytes to allocate for PCache */
	u32 szPageDflt = SQLITE_DEFAULT_PAGE_SIZE;	/* Default page size */
	int nUri = 0;		/* Number of bytes of URI args at *zUri */

	/* Figure out how much space is required for each journal file-handle
	 * (there are two of them, the main journal and the sub-journal).
	 */
	journalFileSize = ROUND8(sqlite3JournalSize(pVfs));

	/* Set the output variable to NULL in case an error occurs. */
	*ppPager = 0;

#ifndef SQLITE_OMIT_MEMORYDB
	if (flags & PAGER_MEMORY) {
		memDb = 1;
		if (zFilename && zFilename[0]) {
			zPathname = sqlite3DbStrDup(0, zFilename);
			if (zPathname == 0)
				return SQLITE_NOMEM_BKPT;
			nPathname = sqlite3Strlen30(zPathname);
			zFilename = 0;
		}
	}
#endif

	/* Allocate memory for the Pager structure, PCache object, the
	 * three file descriptors, the database file name and the journal
	 * file name. The layout in memory is as follows:
	 *
	 *     Pager object                    (sizeof(Pager) bytes)
	 *     PCache object                   (sqlite3PcacheSize() bytes)
	 *     Database file handle            (pVfs->szOsFile bytes)
	 *     Sub-journal file handle         (journalFileSize bytes)
	 *     Main journal file handle        (journalFileSize bytes)
	 *     Database file name              (nPathname+1 bytes)
	 *     Journal file name               (nPathname+8+1 bytes)
	 */
	pPtr = (u8 *) sqlite3MallocZero(ROUND8(sizeof(*pPager)) +	/* Pager structure */
					ROUND8(pcacheSize) +	/* PCache object */
					ROUND8(pVfs->szOsFile) +	/* The main db file */
					journalFileSize * 2 +	/* The two journal files */
					nPathname + 1 + nUri +	/* zFilename */
					nPathname + 8 + 2	/* zJournal */
	    );
	assert(EIGHT_BYTE_ALIGNMENT(SQLITE_INT_TO_PTR(journalFileSize)));
	if (!pPtr) {
		sqlite3DbFree(0, zPathname);
		return SQLITE_NOMEM_BKPT;
	}
	pPager = (Pager *) (pPtr);
	pPager->pPCache = (PCache *) (pPtr += ROUND8(sizeof(*pPager)));
	pPager->fd = (sqlite3_file *) (pPtr += ROUND8(pcacheSize));
	pPager->sjfd = (sqlite3_file *) (pPtr += ROUND8(pVfs->szOsFile));
	pPager->jfd = (sqlite3_file *) (pPtr += journalFileSize);
	pPager->zFilename = (char *)(pPtr += journalFileSize);
	assert(EIGHT_BYTE_ALIGNMENT(pPager->jfd));
	pPager->pVfs = pVfs;
	pPager->vfsFlags = vfsFlags;

	/* If a temporary file is requested, it is not opened immediately.
	 * In this case we accept the default page size and delay actually
	 * opening the file until the first call to OsWrite().
	 *
	 * This branch is also run for an in-memory database. An in-memory
	 * database is the same as a temp-file that is never written out to
	 * disk and uses an in-memory rollback journal.
	 *
	 * This branch also runs for files marked as immutable.
	 */
	tempFile = 1;
	pPager->eState = PAGER_READER;	/* Pretend we already have a lock */
	pPager->eLock = EXCLUSIVE_LOCK;	/* Pretend we are in EXCLUSIVE mode */
	pPager->noLock = 1;	/* Do no locking */
	readOnly = (vfsFlags & SQLITE_OPEN_READONLY);

	/* The following call to PagerSetPagesize() serves to set the value of
	 * Pager.pageSize and to allocate the Pager.pTmpSpace buffer.
	 */
	if (rc == SQLITE_OK) {
		assert(pPager->memDb == 0);
		rc = sqlite3PagerSetPagesize(pPager, &szPageDflt, -1);
		testcase(rc != SQLITE_OK);
	}

	/* Initialize the PCache object. */
	if (rc == SQLITE_OK) {
		nExtra = ROUND8(nExtra);
		assert(nExtra >= 8 && nExtra < 1000);
		rc = sqlite3PcacheOpen(szPageDflt, nExtra, !memDb,
				       0, (void *)pPager, pPager->pPCache);
	}

	/* If an error occurred above, free the  Pager structure and close the file.
	 */
	if (rc != SQLITE_OK) {
		sqlite3OsClose(pPager->fd);
		sqlite3PageFree(pPager->pTmpSpace);
		sqlite3_free(pPager);
		return rc;
	}

	PAGERTRACE(("OPEN %d %s\n", FILEHANDLEID(pPager->fd),
		    pPager->zFilename));
	IOTRACE(("OPEN %p %s\n", pPager, pPager->zFilename))

	    pPager->useJournal = (u8) useJournal;
	/* pPager->stmtOpen = 0; */
	/* pPager->stmtInUse = 0; */
	/* pPager->nRef = 0; */
	/* pPager->stmtSize = 0; */
	/* pPager->stmtJSize = 0; */
	/* pPager->nPage = 0; */
	pPager->mxPgno = SQLITE_MAX_PAGE_COUNT;
	/* pPager->state = PAGER_UNLOCK; */
	/* pPager->errMask = 0; */
	pPager->tempFile = (u8) tempFile;
	assert(tempFile == PAGER_LOCKINGMODE_NORMAL
	       || tempFile == PAGER_LOCKINGMODE_EXCLUSIVE);
	assert(PAGER_LOCKINGMODE_EXCLUSIVE == 1);
	pPager->exclusiveMode = (u8) tempFile;
	pPager->changeCountDone = pPager->tempFile;
	pPager->memDb = (u8) memDb;
	pPager->readOnly = (u8) readOnly;
	assert(useJournal || pPager->tempFile);
	pPager->noSync = pPager->tempFile;
	if (pPager->noSync) {
		assert(pPager->fullSync == 0);
		assert(pPager->extraSync == 0);
		assert(pPager->syncFlags == 0);
		assert(pPager->walSyncFlags == 0);
		assert(pPager->ckptSyncFlags == 0);
	} else {
		pPager->fullSync = 1;
		pPager->extraSync = 0;
		pPager->syncFlags = SQLITE_SYNC_NORMAL;
		pPager->ckptSyncFlags = SQLITE_SYNC_NORMAL;
	}
	/* pPager->pFirst = 0; */
	/* pPager->pFirstSynced = 0; */
	/* pPager->pLast = 0; */
	pPager->nExtra = (u16) nExtra;
	pPager->journalSizeLimit = SQLITE_DEFAULT_JOURNAL_SIZE_LIMIT;
	assert(isOpen(pPager->fd) || tempFile);
	setSectorSize(pPager);
	if (!useJournal) {
		pPager->journalMode = PAGER_JOURNALMODE_OFF;
	} else if (memDb) {
		pPager->journalMode = PAGER_JOURNALMODE_MEMORY;
	}
	/* pPager->xBusyHandler = 0; */
	/* pPager->pBusyHandlerArg = 0; */
	pPager->xReiniter = xReinit;
	setGetterMethod(pPager);
	/* memset(pPager->aHash, 0, sizeof(pPager->aHash)); */
	/* pPager->szMmap = SQLITE_DEFAULT_MMAP_SIZE // will be set by btree.c */

	*ppPager = pPager;
	return SQLITE_OK;
}

/*
 * This function is called to obtain a shared lock on the database file.
 * It is illegal to call sqlite3PagerGet() until after this function
 * has been successfully called. If a shared-lock is already held when
 * this function is called, it is a no-op.
 *
 * The following operations are also performed by this function.
 *
 *   1) If the pager is currently in PAGER_OPEN state (no lock held
 *      on the database file), then an attempt is made to obtain a
 *      SHARED lock on the database file. Immediately after obtaining
 *      the SHARED lock, the file-system is checked for a hot-journal,
 *      which is played back if present. Following any hot-journal
 *      rollback, the contents of the cache are validated by checking
 *      the 'change-counter' field of the database file header and
 *      discarded if they are found to be invalid.
 *
 *   2) If the pager is running in exclusive-mode, and there are currently
 *      no outstanding references to any pages, and is in the error state,
 *      then an attempt is made to clear the error state by discarding
 *      the contents of the page cache and rolling back any open journal
 *      file.
 *
 * If everything is successful, SQLITE_OK is returned. If an IO error
 * occurs while locking the database, checking for a hot-journal file or
 * rolling back a journal file, the IO error code is returned.
 */
int
sqlite3PagerSharedLock(Pager * pPager)
{
	int rc = SQLITE_OK;	/* Return code */

	/* This routine is only called from b-tree and only when there are no
	 * outstanding pages. This implies that the pager state should either
	 * be OPEN or READER. READER is only possible if the pager is or was in
	 * exclusive access mode.
	 */
	assert(sqlite3PcacheRefCount(pPager->pPCache) == 0);
	assert(assert_pager_state(pPager));
	assert(pPager->errCode == SQLITE_OK);

	if (pagerUseWal(pPager)) {
		assert(rc == SQLITE_OK);
		rc = pagerBeginReadTransaction(pPager);
	}

	pPager->eState = PAGER_READER;
	pPager->hasHeldSharedLock = 1;
	return rc;
}

/*
 * The page getter methods each try to acquire a reference to a
 * page with page number pgno. If the requested reference is
 * successfully obtained, it is copied to *ppPage and SQLITE_OK returned.
 *
 * There are different implementations of the getter method depending
 * on the current state of the pager.
 *
 *     getPageNormal()         --  The normal getter
 *     getPageError()          --  Used if the pager is in an error state
 *     getPageMmap()           --  Used if memory-mapped I/O is enabled
 *
 * If the requested page is already in the cache, it is returned.
 * Otherwise, a new page object is allocated and populated with data
 * read from the database file. In some cases, the pcache module may
 * choose not to allocate a new page object and may reuse an existing
 * object with no outstanding references.
 *
 * The extra data appended to a page is always initialized to zeros the
 * first time a page is loaded into memory. If the page requested is
 * already in the cache when this function is called, then the extra
 * data is left as it was when the page object was last used.
 *
 * If the database image is smaller than the requested page or if
 * the flags parameter contains the PAGER_GET_NOCONTENT bit and the
 * requested page is not already stored in the cache, then no
 * actual disk read occurs. In this case the memory image of the
 * page is initialized to all zeros.
 *
 * If PAGER_GET_NOCONTENT is true, it means that we do not care about
 * the contents of the page. This occurs in two scenarios:
 *
 *   a) When reading a free-list leaf page from the database, and
 *
 *   b) When a savepoint is being rolled back and we need to load
 *      a new page into the cache to be filled with the data read
 *      from the savepoint journal.
 *
 * If PAGER_GET_NOCONTENT is true, then the data returned is zeroed instead
 * of being read from the database. Additionally, the bits corresponding
 * to pgno in Pager.pInJournal (bitvec of pages already written to the
 * journal file) and the PagerSavepoint.pInSavepoint bitvecs of any open
 * savepoints are set. This means if the page is made writable at any
 * point in the future, using a call to sqlite3PagerWrite(), its contents
 * will not be journaled. This saves IO.
 *
 * The acquisition might fail for several reasons.  In all cases,
 * an appropriate error code is returned and *ppPage is set to NULL.
 *
 * See also sqlite3PagerLookup().  Both this routine and Lookup() attempt
 * to find a page in the in-memory cache first.  If the page is not already
 * in memory, this routine goes to disk to read it in whereas Lookup()
 * just returns 0.  This routine acquires a read-lock the first time it
 * has to go to disk, and could also playback an old journal if necessary.
 * Since Lookup() never goes to disk, it never has to deal with locks
 * or journal files.
 */
static int
getPageNormal(Pager * pPager,	/* The pager open on the database file */
	      Pgno pgno,	/* Page number to fetch */
	      DbPage ** ppPage,	/* Write a pointer to the page here */
	      int flags		/* PAGER_GET_XXX flags */
    )
{
	int rc;
	PgHdr *pPg;
	u8 noContent;		/* True if PAGER_GET_NOCONTENT is set */
	sqlite3_pcache_page *pBase;

	assert(pPager->errCode == SQLITE_OK);
	assert(pPager->eState >= PAGER_READER);
	assert(assert_pager_state(pPager));
	assert(pPager->hasHeldSharedLock == 1);

	if (pgno == 0)
		return SQLITE_CORRUPT_BKPT;
	pBase = sqlite3PcacheFetch(pPager->pPCache, pgno, 3);
	if (pBase == 0) {
		pPg = 0;
		rc = sqlite3PcacheFetchStress(pPager->pPCache, pgno, &pBase);
		if (rc != SQLITE_OK)
			goto pager_acquire_err;
		if (pBase == 0) {
			rc = SQLITE_NOMEM_BKPT;
			goto pager_acquire_err;
		}
	}
	pPg = *ppPage = sqlite3PcacheFetchFinish(pPager->pPCache, pgno, pBase);
	assert(pPg == (*ppPage));
	assert(pPg->pgno == pgno);
	assert(pPg->pPager == pPager || pPg->pPager == 0);

	noContent = (flags & PAGER_GET_NOCONTENT) != 0;
	if (pPg->pPager && !noContent) {
		/* In this case the pcache already contains an initialized copy of
		 * the page. Return without further ado.
		 */
		assert(pgno <= PAGER_MAX_PGNO && pgno != PAGER_MJ_PGNO(pPager));
		pPager->aStat[PAGER_STAT_HIT]++;
		return SQLITE_OK;

	} else {
		/* The pager cache has created a new page. Its content needs to
		 * be initialized. But first some error checks:
		 *
		 * (1) The maximum page number is 2^31
		 * (2) Never try to fetch the locking page
		 */
		if (pgno > PAGER_MAX_PGNO || pgno == PAGER_MJ_PGNO(pPager)) {
			rc = SQLITE_CORRUPT_BKPT;
			goto pager_acquire_err;
		}

		pPg->pPager = pPager;

		assert(!isOpen(pPager->fd) || !MEMDB);
		if (!isOpen(pPager->fd) || pPager->dbSize < pgno || noContent) {
			if (pgno > pPager->mxPgno) {
				rc = SQLITE_FULL;
				goto pager_acquire_err;
			}
			if (noContent) {
				/* Failure to set the bits in the InJournal bit-vectors is benign.
				 * It merely means that we might do some extra work to journal a
				 * page that does not need to be journaled.  Nevertheless, be sure
				 * to test the case where a malloc error occurs while trying to set
				 * a bit in a bit vector.
				 */
				sqlite3BeginBenignMalloc();
				if (pgno <= pPager->dbOrigSize) {
					sqlite3BitvecSet(pPager->pInJournal,
							 pgno);
					testcase(rc == SQLITE_NOMEM);
				}
				testcase(rc == SQLITE_NOMEM);
				sqlite3EndBenignMalloc();
			}
			memset(pPg->pData, 0, pPager->pageSize);
			IOTRACE(("ZERO %p %d\n", pPager, pgno));
		}
		pager_set_pagehash(pPg);
	}
	return SQLITE_OK;

 pager_acquire_err:
	assert(rc != SQLITE_OK);
	if (pPg) {
		sqlite3PcacheDrop(pPg);
	}
	*ppPage = 0;
	return rc;
}

/* The page getter method for when the pager is an error state */
static int
getPageError(Pager * pPager,	/* The pager open on the database file */
	     Pgno pgno,		/* Page number to fetch */
	     DbPage ** ppPage,	/* Write a pointer to the page here */
	     int flags		/* PAGER_GET_XXX flags */
    )
{
	UNUSED_PARAMETER(pgno);
	UNUSED_PARAMETER(flags);
	assert(pPager->errCode != SQLITE_OK);
	*ppPage = 0;
	return pPager->errCode;
}

/* Dispatch all page fetch requests to the appropriate getter method.
 */
int
sqlite3PagerGet(Pager * pPager,	/* The pager open on the database file */
		Pgno pgno,	/* Page number to fetch */
		DbPage ** ppPage,	/* Write a pointer to the page here */
		int flags	/* PAGER_GET_XXX flags */
    )
{
	return pPager->xGet(pPager, pgno, ppPage, flags);
}

/*
 * Acquire a page if it is already in the in-memory cache.  Do
 * not read the page from disk.  Return a pointer to the page,
 * or 0 if the page is not in cache.
 *
 * See also sqlite3PagerGet().  The difference between this routine
 * and sqlite3PagerGet() is that _get() will go to the disk and read
 * in the page if the page is not already in cache.  This routine
 * returns NULL if the page is not in cache or if a disk I/O error
 * has ever happened.
 */
DbPage *
sqlite3PagerLookup(Pager * pPager, Pgno pgno)
{
	sqlite3_pcache_page *pPage;
	assert(pPager != 0);
	assert(pgno != 0);
	assert(pPager->pPCache != 0);
	pPage = sqlite3PcacheFetch(pPager->pPCache, pgno, 0);
	assert(pPage == 0 || pPager->hasHeldSharedLock);
	if (pPage == 0)
		return 0;
	return sqlite3PcacheFetchFinish(pPager->pPCache, pgno, pPage);
}

/*
 * Release a page reference.
 *
 * If the number of references to the page drop to zero, then the
 * page is added to the LRU list.  When all references to all pages
 * are released, a rollback occurs and the lock on the database is
 * removed.
 */
void
sqlite3PagerUnrefNotNull(DbPage * pPg)
{
	assert(pPg != 0);
	sqlite3PcacheRelease(pPg);
}

void
sqlite3PagerUnref(DbPage * pPg)
{
	if (pPg)
		sqlite3PagerUnrefNotNull(pPg);
}

/*
 * Mark a single data page as writeable. The page is written into the
 * main journal or sub-journal as required. If the page is written into
 * one of the journals, the corresponding bit is set in the
 * Pager.pInJournal bitvec and the PagerSavepoint.pInSavepoint bitvecs
 * of any open savepoints as appropriate.
 */
static int
pager_write(PgHdr * pPg)
{
	Pager *pPager = pPg->pPager;
	int rc = SQLITE_OK;

	/* This routine is not called unless a write-transaction has already
	 * been started. The journal file may or may not be open at this point.
	 * It is never called in the ERROR state.
	 */
	assert(assert_pager_state(pPager));
	assert(pPager->errCode == 0);
	assert(pPager->readOnly == 0);
	CHECK_PAGE(pPg);
	assert(assert_pager_state(pPager));

	/* Mark the page that is about to be modified as dirty. */
	sqlite3PcacheMakeDirty(pPg);

	/* If a rollback journal is in use, them make sure the page that is about
	 * to change is in the rollback journal, or if the page is a new page off
	 * then end of the file, make sure it is marked as PGHDR_NEED_SYNC.
	 */
	assert((pPager->pInJournal != 0) == isOpen(pPager->jfd));

	/* The PGHDR_DIRTY bit is set above when the page was added to the dirty-list
	 * and before writing the page into the rollback journal.  Wait until now,
	 * after the page has been successfully journalled, before setting the
	 * PGHDR_WRITEABLE bit that indicates that the page can be safely modified.
	 */
	pPg->flags |= PGHDR_WRITEABLE;

	/* Update the database size and return. */
	if (pPager->dbSize < pPg->pgno) {
		pPager->dbSize = pPg->pgno;
	}
	return rc;
}

/*
 * Mark a data page as writeable. This routine must be called before
 * making changes to a page. The caller must check the return value
 * of this function and be careful not to change any page data unless
 * this routine returns SQLITE_OK.
 *
 * The difference between this function and pager_write() is that this
 * function also deals with the special case where 2 or more pages
 * fit on a single disk sector. In this case all co-resident pages
 * must have been written to the journal file before returning.
 *
 * If an error occurs, SQLITE_NOMEM or an IO error code is returned
 * as appropriate. Otherwise, SQLITE_OK.
 */
int
sqlite3PagerWrite(PgHdr * pPg)
{
	Pager *pPager = pPg->pPager;
	assert((pPg->flags & PGHDR_MMAP) == 0);
	assert(assert_pager_state(pPager));
	if ((pPg->flags & PGHDR_WRITEABLE) != 0 && pPager->dbSize >= pPg->pgno) {
		return SQLITE_OK;
	} else if (pPager->errCode) {
		return pPager->errCode;
	} else {
		return pager_write(pPg);
	}
}

/*
 * Return TRUE if the page given in the argument was previously passed
 * to sqlite3PagerWrite().  In other words, return TRUE if it is ok
 * to change the content of the page.
 */
#ifndef NDEBUG
int
sqlite3PagerIswriteable(DbPage * pPg)
{
	return pPg->flags & PGHDR_WRITEABLE;
}
#endif

/*
 * A call to this routine tells the pager that it is not necessary to
 * write the information on page pPg back to the disk, even though
 * that page might be marked as dirty.  This happens, for example, when
 * the page has been added as a leaf of the freelist and so its
 * content no longer matters.
 *
 * The overlying software layer calls this routine when all of the data
 * on the given page is unused. The pager marks the page as clean so
 * that it does not get written to disk.
 *
 * Tests show that this optimization can quadruple the speed of large
 * DELETE operations.
 *
 * This optimization cannot be used with a temp-file, as the page may
 * have been dirty at the start of the transaction. In that case, if
 * memory pressure forces page pPg out of the cache, the data does need
 * to be written out to disk so that it may be read back in if the
 * current transaction is rolled back.
 */
void
sqlite3PagerDontWrite(PgHdr * pPg)
{
	Pager *pPager = pPg->pPager;
	if (!pPager->tempFile && (pPg->flags & PGHDR_DIRTY)
	    && pPager->nSavepoint == 0) {
		PAGERTRACE(("DONT_WRITE page %d of %d\n", pPg->pgno,
			    PAGERID(pPager)));
		IOTRACE(("CLEAN %p %d\n", pPager, pPg->pgno))
		    pPg->flags |= PGHDR_DONT_WRITE;
		pPg->flags &= ~PGHDR_WRITEABLE;
		testcase(pPg->flags & PGHDR_NEED_SYNC);
		pager_set_pagehash(pPg);
	}
}

/*
 * This function may only be called while a write-transaction is active in
 * rollback. If the connection is in WAL mode, this call is a no-op.
 * Otherwise, if the connection does not already have an EXCLUSIVE lock on
 * the database file, an attempt is made to obtain one.
 *
 * If the EXCLUSIVE lock is already held or the attempt to obtain it is
 * successful, or the connection is in WAL mode, SQLITE_OK is returned.
 * Otherwise, either SQLITE_BUSY or an SQLITE_IOERR_XXX error code is
 * returned.
 */
int
sqlite3PagerExclusiveLock(Pager * pPager)
{
	int rc = pPager->errCode;
	assert(assert_pager_state(pPager));
	return rc;
}

/*
 * Sync the database file for the pager pPager. zMaster points to the name
 * of a master journal file that should be written into the individual
 * journal file. zMaster may be NULL, which is interpreted as no master
 * journal (a single database transaction).
 *
 * This routine ensures that:
 *
 *   * The database file change-counter is updated,
 *   * the journal is synced (unless the atomic-write optimization is used),
 *   * all dirty pages are written to the database file,
 *   * the database file is truncated (if required), and
 *   * the database file synced.
 *
 * The only thing that remains to commit the transaction is to finalize
 * (delete, truncate or zero the first part of) the journal file (or
 * delete the master journal file if specified).
 *
 * Note that if zMaster==NULL, this does not overwrite a previous value
 * passed to an sqlite3PagerCommitPhaseOne() call.
 *
 * If the final parameter - noSync - is true, then the database file itself
 * is not synced. The caller must call sqlite3PagerSync() directly to
 * sync the database file before calling CommitPhaseTwo() to delete the
 * journal file in this case.
 */
int
sqlite3PagerCommitPhaseOne(Pager * pPager	/* Pager object */
    )
{
	int rc = SQLITE_OK;	/* Return code */

	assert(assert_pager_state(pPager));

	/* If a prior error occurred, report that error again. */
	if (NEVER(pPager->errCode))
		return pPager->errCode;

	/* Provide the ability to easily simulate an I/O error during testing */
	if (sqlite3FaultSim(400))
		return SQLITE_IOERR;

	PAGERTRACE(("DATABASE SYNC: File=%s zMaster=%s nSize=%d\n",
		    pPager->zFilename, zMaster, pPager->dbSize));

	/* If no database changes have been made, return early. */
	if (pPager->eState < PAGER_WRITER_CACHEMOD)
		return SQLITE_OK;

	assert(MEMDB == 0 || pPager->tempFile);
	assert(isOpen(pPager->fd) || pPager->tempFile);
	if (rc == SQLITE_OK && !pagerUseWal(pPager)) {
		pPager->eState = PAGER_WRITER_FINISHED;
	}
	return rc;
}

/*
 * Return TRUE if the database file is opened read-only.  Return FALSE
 * if the database is (in theory) writable.
 */
u8
sqlite3PagerIsreadonly(Pager * pPager)
{
	return pPager->readOnly;
}

#ifdef SQLITE_DEBUG
/*
 * Return the sum of the reference counts for all pages held by pPager.
 */
int
sqlite3PagerRefcount(Pager * pPager)
{
	return sqlite3PcacheRefCount(pPager->pPCache);
}
#endif

/*
 * Return the number of references to the specified page.
 */
int
sqlite3PagerPageRefcount(DbPage * pPage)
{
	return sqlite3PcachePageRefcount(pPage);
}

/*
 * This function is called to rollback or release (commit) a savepoint.
 * The savepoint to release or rollback need not be the most recently
 * created savepoint.
 *
 * Parameter op is always either SAVEPOINT_ROLLBACK or SAVEPOINT_RELEASE.
 * If it is SAVEPOINT_RELEASE, then release and destroy the savepoint with
 * index iSavepoint. If it is SAVEPOINT_ROLLBACK, then rollback all changes
 * that have occurred since the specified savepoint was created.
 *
 * The savepoint to rollback or release is identified by parameter
 * iSavepoint. A value of 0 means to operate on the outermost savepoint
 * (the first created). A value of (Pager.nSavepoint-1) means operate
 * on the most recently created savepoint. If iSavepoint is greater than
 * (Pager.nSavepoint-1), then this function is a no-op.
 *
 * If a negative value is passed to this function, then the current
 * transaction is rolled back. This is different to calling
 * sqlite3PagerRollback() because this function does not terminate
 * the transaction or unlock the database, it just restores the
 * contents of the database to its original state.
 *
 * In any case, all savepoints with an index greater than iSavepoint
 * are destroyed. If this is a release operation (op==SAVEPOINT_RELEASE),
 * then savepoint iSavepoint is also destroyed.
 *
 * This function may return SQLITE_NOMEM if a memory allocation fails,
 * or an IO error code if an IO error occurs while rolling back a
 * savepoint. If no errors occur, SQLITE_OK is returned.
 */
int
sqlite3PagerSavepoint(Pager * pPager, int op, int iSavepoint)
{
	int rc = pPager->errCode;

#ifdef SQLITE_ENABLE_ZIPVFS
	if (op == SAVEPOINT_RELEASE)
		rc = SQLITE_OK;
#endif

	assert(op == SAVEPOINT_RELEASE || op == SAVEPOINT_ROLLBACK);
	assert(iSavepoint >= 0 || op == SAVEPOINT_ROLLBACK);

	if (rc == SQLITE_OK && iSavepoint < pPager->nSavepoint) {
		int ii;		/* Iterator variable */
		int nNew;	/* Number of remaining savepoints after this op. */

		/* Figure out how many savepoints will still be active after this
		 * operation. Store this value in nNew. Then free resources associated
		 * with any savepoints that are destroyed by this operation.
		 */
		nNew = iSavepoint + ((op == SAVEPOINT_RELEASE) ? 0 : 1);
		for (ii = nNew; ii < pPager->nSavepoint; ii++) {
			sqlite3BitvecDestroy(pPager->aSavepoint[ii].
					     pInSavepoint);
		}
		pPager->nSavepoint = nNew;

		/* If this is a release of the outermost savepoint, truncate
		 * the sub-journal to zero bytes in size.
		 */
		if (op == SAVEPOINT_RELEASE) {
			if (nNew == 0 && isOpen(pPager->sjfd)) {
				/* Only truncate if it is an in-memory sub-journal. */
				if (sqlite3JournalIsInMemory(pPager->sjfd)) {
					rc = sqlite3OsTruncate(pPager->sjfd, 0);
					assert(rc == SQLITE_OK);
				}
				pPager->nSubRec = 0;
			}
		}
#ifdef SQLITE_ENABLE_ZIPVFS
		/* If the cache has been modified but the savepoint cannot be rolled
		 * back journal_mode=off, put the pager in the error state. This way,
		 * if the VFS used by this pager includes ZipVFS, the entire transaction
		 * can be rolled back at the ZipVFS level.
		 */
		else if (pPager->journalMode == PAGER_JOURNALMODE_OFF
			 && pPager->eState >= PAGER_WRITER_CACHEMOD) {
			pPager->errCode = SQLITE_ABORT;
			pPager->eState = PAGER_ERROR;
			setGetterMethod(pPager);
		}
#endif
	}

	return rc;
}

/*
 * Return the full pathname of the database file.
 *
 * Except, if the pager is in-memory only, then return an empty string if
 * nullIfMemDb is true.  This routine is called with nullIfMemDb==1 when
 * used to report the filename to the user, for compatibility with legacy
 * behavior.  But when the Btree needs to know the filename for matching to
 * shared cache, it uses nullIfMemDb==0 so that in-memory databases can
 * participate in shared-cache.
 */
const char *
sqlite3PagerFilename(Pager * pPager, int nullIfMemDb)
{
	return (nullIfMemDb && pPager->memDb) ? "" : pPager->zFilename;
}

/*
 * Return the file handle for the database file associated
 * with the pager.  This might return NULL if the file has
 * not yet been opened.
 */
sqlite3_file *
sqlite3PagerFile(Pager * pPager)
{
	return pPager->fd;
}

/*
 * Return the file handle for the journal file (if it exists).
 * This will be either the rollback journal or the WAL file.
 */
sqlite3_file *
sqlite3PagerJrnlFile(Pager * pPager)
{
	return pPager->jfd;
}

#ifdef SQLITE_HAS_CODEC
/*
 * Set or retrieve the codec for this pager
 */
void
sqlite3PagerSetCodec(Pager * pPager,
		     void *(*xCodec) (void *, void *, Pgno, int),
		     void (*xCodecSizeChng) (void *, int, int),
		     void (*xCodecFree) (void *), void *pCodec)
{
	if (pPager->xCodecFree)
		pPager->xCodecFree(pPager->pCodec);
	pPager->xCodec = pPager->memDb ? 0 : xCodec;
	pPager->xCodecSizeChng = xCodecSizeChng;
	pPager->xCodecFree = xCodecFree;
	pPager->pCodec = pCodec;
	setGetterMethod(pPager);
	pagerReportSize(pPager);
}

void *
sqlite3PagerGetCodec(Pager * pPager)
{
	return pPager->pCodec;
}

/*
 * This function is called by the wal module when writing page content
 * into the log file.
 *
 * This function returns a pointer to a buffer containing the encrypted
 * page content. If a malloc fails, this function may return NULL.
 */
void *
sqlite3PagerCodec(PgHdr * pPg)
{
	void *aData = 0;
	CODEC2(pPg->pPager, pPg->pData, pPg->pgno, 6, return 0, aData);
	return aData;
}

/*
 * Return the current pager state
 */
int
sqlite3PagerState(Pager * pPager)
{
	return pPager->eState;
}
#endif				/* SQLITE_HAS_CODEC */

/*
 * The page handle passed as the first argument refers to a dirty page
 * with a page number other than iNew. This function changes the page's
 * page number to iNew and sets the value of the PgHdr.flags field to
 * the value passed as the third parameter.
 */
void
sqlite3PagerRekey(DbPage * pPg, Pgno iNew, u16 flags)
{
	assert(pPg->pgno != iNew);
	pPg->flags = flags;
	sqlite3PcacheMove(pPg, iNew);
}

/*
 * Return a pointer to the data for the specified page.
 */
void *
sqlite3PagerGetData(DbPage * pPg)
{
	assert(pPg->nRef > 0 || pPg->pPager->memDb);
	return pPg->pData;
}

/*
 * Return a pointer to the Pager.nExtra bytes of "extra" space
 * allocated along with the specified page.
 */
void *
sqlite3PagerGetExtra(DbPage * pPg)
{
	return pPg->pExtra;
}

/*
 * Return the current journal mode.
 */
int
sqlite3PagerGetJournalMode(Pager * pPager)
{
	return (int)pPager->journalMode;
}

#ifdef SQLITE_ENABLE_ZIPVFS
/*
 * A read-lock must be held on the pager when this function is called. If
 * the pager is in WAL mode and the WAL file currently contains one or more
 * frames, return the size in bytes of the page images stored within the
 * WAL frames. Otherwise, if this is not a WAL database or the WAL file
 * is empty, return 0.
 */
int
sqlite3PagerWalFramesize(Pager * pPager)
{
	assert(pPager->eState >= PAGER_READER);
	return sqlite3WalFramesize(pPager->pWal);
}
#endif

#endif				/* SQLITE_OMIT_DISKIO */
